;; Generated package description from with-simulated-input.el  -*- no-byte-compile: t -*-
(define-package "with-simulated-input" "3.0" "A macro to simulate user input non-interactively" '((emacs "24.4")) :commit "07bdfbe9ab9eab4a04ad02e98b011649a4f4e6a2" :authors '(("Ryan C. Thompson" . "rct@thompsonclan.org") ("Nikita Bloshchanevich" . "nikblos@outlook.com")) :maintainer '("Ryan C Thompson" . "rct@thompsonclan.org") :keywords '("lisp" "tools" "extensions") :url "https://github.com/DarwinAwardWinner/with-simulated-input")
